/* nicejob - renice jobs having used more then a specified user time
   All rights granted. All warranties voided
   Chun-Chung Chen <cjj@u.washington.edu> */

/* Options:
   -t, --time
   -n, --nice
   -h, --help
   -v, --version
*/

#define PROGRAM_NAME "nicejob"
#define AUTHORS "Chun-Chung Chen"
#define VERSION "0.2"
char *program_name;

#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <getopt.h>

/* default values */
long timelimit = 10; /* minute(s) */
int nicelevel = 10;
uid_t firstuid = 100;

#define BUFSIZE 256

static struct option const long_options[] = {
	{"time", required_argument, NULL, 't'},
	{"nice", required_argument, NULL, 'n'},
	{"firstuid", required_argument, NULL, 'f'},
	{"help", no_argument, NULL, 250},
	{"version", no_argument, NULL, 251},
	{NULL, 0, NULL, 0}
};

void usage(int status) {
	if (status != 0) {
		fprintf(stderr, "Try `%s --help' for more information.\n", program_name);
		exit(status);
	}
	printf("\
Usage: %s [OPTION]...\n\
Renice jobs having used more then a specified CPU time.\n\
  -t, --time=TIME    set the time limit for a job to be reniced to TIME(min)\n\
                     default=%ld\n\
  -n, --nice=VALUE   attempt to set the priority of jobs to VALUE\n\
                     default=%d\n\
  -f, --firstuid=UID set the lowest uid for the processes to be reniced to UID\n\
                     default=%d\n\
      --help              display this help and exit\n\
      --version           output version information an exit\n",
	       program_name, timelimit, nicelevel, firstuid);
	exit(status);
}

int main(int argc, char **argv) {
	int optc;
	char buf[BUFSIZE];
	char *pname;
	DIR *proc;	
	struct dirent *e;
	struct stat pstat;
	FILE *st;
	long read = 0;
	long utime;
	long stime;
	int pr;
	int np;
	int i;
	int job_niced = 0;
	pid_t pid;
	uid_t uid;

	program_name = argv[0];

	while((optc = getopt_long(argc, argv, "t:n:f:", long_options, NULL)) != -1) switch (optc) {
	case 0:
		break;
	case 't':
		timelimit = strtol(optarg, 0, 0);
		break;
	case 'n':
		nicelevel = strtol(optarg, 0, 0);
		break;
	case 'f':
		firstuid = strtol(optarg, 0, 0);
		break;
	case 250:
		usage(0);
		break;
	case 251:
		printf("%s Version %s\n", PROGRAM_NAME, VERSION);
		printf("Written by %s.\n", AUTHORS);
	default:
		usage(1);
	}
	if ((proc = opendir("/proc")) == NULL) {
		fprintf(stderr, "can't open /proc!");
		return -1;
	}
	/* for each subdirs with process number as name */
	while ((e = readdir(proc)) != 0) {
		if (!isdigit(e->d_name[0])) continue; /* not a process directory */
		snprintf(buf, BUFSIZE, "/proc/%s/stat", e->d_name);
		if (stat(buf, &pstat) != 0) continue; /* can't stat the stat file */
		uid = pstat.st_uid;
		if (uid < firstuid) continue; /* low UID job */
		if ((st = fopen(buf, "r")) == NULL) continue; /* can't open stat file */
		read = fread(buf, 1, BUFSIZE, st);
		buf[read] = '\0';
		fclose(st);
		if (read == 0) continue; /* stat file empty ?! */
		/* a walk through the stat file */
		pid = strtol(strtok(buf, " \n"), 0, 0);         /* field 1 */
		pname = strtok(NULL, " \n");                    /* field 2 */
		for (i = 3; i <= 13; i ++) strtok(NULL, " \n"); /* field 3 - 13 */
		utime = strtol(strtok(NULL, " \n"), 0, 0);      /* field 14 */
		stime = strtol(strtok(NULL, " \n"), 0, 0);      /* field 15 */
		if ((utime + stime) / 6000 < timelimit) continue; /* time limit not exceeded */
		pr = getpriority(PRIO_PROCESS, pid);
		if (pr >= nicelevel) continue; /* niced already */
		if (setpriority(PRIO_PROCESS, pid, nicelevel) != 0) continue; /* can't renice it */
		np = getpriority(PRIO_PROCESS, pid);
		if (pr == np) continue; /* priority not changed (nicelevel >= 20?) */
		printf("%u%s by %u used %.1f(min) reniced [%d]->[%d]\n", pid, pname, uid, (utime + stime) / 6e3, pr, np);
		job_niced ++;
	}
	closedir(proc);
	if (job_niced > 0) printf("%d job(s) reniced\n", job_niced);
	return 0;
}
